#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2014-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Do the unit testing for the utils library."""
import contextlib
import logging
import os
import re
import tarfile
import tempfile
import unittest
import warnings
from io import BytesIO

import numpy as np
import pytest
import responses

from pyspectral import utils
from pyspectral.utils import are_instruments_identical, bytes2string, check_and_adjust_instrument_name, np2str

TEST_RSR = {'20': {}, }
TEST_RSR2 = {'20': {}, }
TEST_RSR['20']['det-1'] = {}
TEST_RSR['20']['det-1']['central_wavelength'] = 3.75
TEST_RSR['20']['det-1']['wavelength'] = np.array([
    3.6123999, 3.6163599, 3.6264927, 3.6363862, 3.646468,
    3.6564937, 3.6664478, 3.6765388, 3.6865413, 3.6964585,
    3.7065142, 3.716509, 3.7264658, 3.7364102, 3.7463682,
    3.7563652, 3.7664226, 3.7763396, 3.7863384, 3.7964207,
    3.8063589, 3.8163606, 3.8264089, 3.8364836, 3.8463381,
    3.8563975, 3.8664163, 3.8763755, 3.8864797, 3.8964978,
    3.9064275, 3.9164873, 3.9264729, 3.9364026, 3.9465107,
    3.9535347], dtype='float32')

TEST_RSR['20']['det-1']['response'] = np.array([
    0.01, 0.0118, 0.01987, 0.03226, 0.05028, 0.0849,
    0.16645, 0.33792, 0.59106, 0.81815, 0.96077, 0.92855,
    0.86008, 0.8661, 0.87697, 0.85412, 0.88922, 0.9541,
    0.95687, 0.91037, 0.91058, 0.94256, 0.94719, 0.94808,
    1., 0.92676, 0.67429, 0.44715, 0.27762, 0.14852,
    0.07141, 0.04151, 0.02925, 0.02085, 0.01414, 0.01], dtype='float32')

TEST_RSR2['20'] = TEST_RSR['20']['det-1']

RESULT_RSR = {'20': {}}
RESULT_RSR['20']['det-1'] = {}
RESULT_RSR['20']['det-1']['wavenumber'] = np.array([
    2529.38208008, 2533.8840332, 2540.390625, 2546.81494141,
    2553.30883789, 2559.88354492, 2566.40722656, 2573.02270508,
    2579.72949219, 2586.37451172, 2593.09375, 2599.87548828,
    2606.55371094, 2613.41674805, 2620.29760742, 2627.18286133,
    2634.06005859, 2641.07421875, 2648.06713867, 2655.03930664,
    2662.14794922, 2669.25170898, 2676.36572266, 2683.50805664,
    2690.69702148, 2697.95288086, 2705.29199219, 2712.56982422,
    2719.94970703, 2727.43554688, 2734.8605957, 2742.38012695,
    2749.9831543, 2757.48510742, 2765.21142578, 2768.24291992],
    dtype='float32')

RESULT_RSR['20']['det-1']['response'] = np.array([
    0.01, 0.01414, 0.02085, 0.02925, 0.04151, 0.07141,
    0.14852, 0.27762, 0.44715, 0.67429, 0.92676, 1.,
    0.94808, 0.94719, 0.94256, 0.91058, 0.91037, 0.95687,
    0.9541, 0.88922, 0.85412, 0.87697, 0.8661, 0.86008,
    0.92855, 0.96077, 0.81815, 0.59106, 0.33792, 0.16645,
    0.0849, 0.05028, 0.03226, 0.01987, 0.0118, 0.01],
    dtype='float32')


class _MockedRSR:
    def __init__(self, usedet=True):
        self.instrument = 'test_sensor'
        if usedet:
            self.tmp_rsr = TEST_RSR
        else:
            self.tmp_rsr = TEST_RSR2
        self.output_dir = tempfile.mkdtemp()

    def __call__(self, chan, plat):
        self.rsr = self.tmp_rsr[chan]
        return self


class RsrTestData(object):
    """Container for the RSR test datasets."""

    def __init__(self):
        """Initialize the class instance."""
        self.rsr = {}
        channel_names = ['ch12', 'ch13', 'ch10', 'ch11', 'ch16', 'ch14',
                         'ch15', 'ch1', 'ch2', 'ch3', 'ch4', 'ch5', 'ch6',
                         'ch7', 'ch8', 'ch9']
        wvl = [9.6372744012936646, 10.407492196078628, 7.3468642293967275,
               8.5926867614178715, 13.280724258676756, 11.239642285822033,
               12.380741429961382, 0.47063607733748003, 0.5099976405799187,
               0.63914891611559055, 0.85668832355426627, 1.6100814361999056,
               2.2568056299864101, 3.8853663735353847, 6.2428987228916233,
               6.9411756334211789]
        ch3_wvl = np.array([0.55518544, 0.56779468, 0.58099002, 0.59481323, 0.60931027,
                            0.62453163, 0.64053291, 0.65737575, 0.67512828, 0.69386619,
                            0.71367401])
        ch3_resp = np.array([2.61000005e-05, 1.07899999e-04, 3.26119992e-03,
                             2.90650606e-01, 9.02460396e-01, 9.60878074e-01,
                             9.97266889e-01, 9.94823873e-01, 7.18220174e-01,
                             8.31819978e-03, 9.34999989e-05])

        idx = 0
        for chname in channel_names:
            self.rsr[chname] = {'det-1': {}}
            self.rsr[chname]['det-1']['central_wavelength'] = wvl[idx]
            idx = idx + 1

        chname = 'ch3'
        self.rsr[chname]['det-1']['wavelength'] = ch3_wvl
        self.rsr[chname]['det-1']['response'] = ch3_resp


class TestUtils(unittest.TestCase):
    """Unit testing the utils library functions."""

    def setUp(self):
        """Set up."""
        self.rsr = RsrTestData()

    def test_convert2wavenumber(self):
        """Testing the conversion of rsr from wavelength to wavenumber."""
        newrsr, info = utils.convert2wavenumber(TEST_RSR)
        unit = info['unit']
        self.assertEqual(unit, 'cm-1')
        self.assertTrue('wavenumber' in newrsr['20']['det-1'])
        self.assertFalse('wavelength' in newrsr['20']['det-1'])
        wvn_res = RESULT_RSR['20']['det-1']['wavenumber']
        wvn = newrsr['20']['det-1']['wavenumber']
        self.assertTrue(np.allclose(wvn_res, wvn))

    def test_convert2hdf5(self):
        """Test the conversion utility from original RSR to HDF5."""
        mocked_rsr_nodet = _MockedRSR(usedet=False)
        utils.convert2hdf5(mocked_rsr_nodet, 'Test_SAT', ['20'])
        fname = f'{mocked_rsr_nodet.output_dir}/rsr_test_sensor_Test_SAT.h5'
        self.assertTrue(os.path.exists(fname))
        try:
            os.remove(fname)
        except OSError:
            pass

        mocked_rsr_det = _MockedRSR(usedet=True)
        utils.convert2hdf5(mocked_rsr_det, 'Test_SAT', ['20'], detectors=['det-1'])
        fname = f'{mocked_rsr_det.output_dir}/rsr_test_sensor_Test_SAT.h5'
        self.assertTrue(os.path.exists(fname))
        try:
            os.remove(fname)
        except OSError:
            pass

    def test_get_bandname_from_wavelength(self):
        """Test the right bandname is found provided the wavelength in micro meters."""
        bname = utils.get_bandname_from_wavelength('abi', 0.4, self.rsr.rsr)
        self.assertEqual(bname, 'ch1')
        with self.assertRaises(AttributeError):
            utils.get_bandname_from_wavelength('abi', 0.5, self.rsr.rsr)

        bname = utils.get_bandname_from_wavelength('abi', 0.6, self.rsr.rsr, epsilon=0.05)
        self.assertEqual(bname, 'ch3')
        bname = utils.get_bandname_from_wavelength('abi', 0.7, self.rsr.rsr)
        self.assertEqual(bname, 'ch3')
        bname = utils.get_bandname_from_wavelength('abi', 0.8, self.rsr.rsr)
        self.assertEqual(bname, 'ch4')
        bname = utils.get_bandname_from_wavelength('abi', 1.0, self.rsr.rsr)
        self.assertEqual(bname, None)

        # Multiple bands returned due to large epsilon
        bname = utils.get_bandname_from_wavelength('abi', 11.1, self.rsr.rsr,
                                                   epsilon=1.0, multiple_bands=True)
        self.assertEqual(bname, ['ch13', 'ch14'])

        # uses generic channel mapping where '20' -> 'ch20'
        bandname = utils.get_bandname_from_wavelength('ufo', 3.7, TEST_RSR)
        self.assertEqual(bandname, 'ch20')

        bandname = utils.get_bandname_from_wavelength('ufo', 3.0, TEST_RSR)
        self.assertIsNone(bandname)

    @staticmethod
    def test_sort_data():
        """Test function sorting data into monotonically increasing values."""
        x_vals = np.array([1.0, 5.6, 30., 2.1, 108.2, 57.8, 1e9, 2.1])
        y_vals = np.array([45., 92., 20., 10., 15., 67., 108., 15.])

        x_sorted = np.array([1., 2.1, 5.6, 30, 57.8, 108.2, 1e9])
        y_sorted = np.array([45., 10., 92., 20., 67., 15., 108.])

        x_vals, y_vals = utils.sort_data(x_vals, y_vals)

        np.testing.assert_equal(x_vals, x_sorted)
        np.testing.assert_equal(y_vals, y_sorted)

    def test_get_wave_range(self):
        """Test the function that produces wavelength ranges from an RSR."""
        wvl_range = utils.get_wave_range(self.rsr.rsr['ch3']['det-1'], 0.15)
        expected_range = [0.59481323, 0.6393011276027835, 0.67512828]
        np.testing.assert_allclose(wvl_range, expected_range)

        wvl_range = utils.get_wave_range(self.rsr.rsr['ch3']['det-1'], 0.5)
        expected_range = [0.60931027, 0.6393011276027835, 0.67512828]
        np.testing.assert_allclose(wvl_range, expected_range)


@pytest.mark.parametrize(
    ("input_value", "exp_except"),
    [
        (np.bytes_("hey"), False),
        (np.array([np.bytes_("hey")]), False),
        (np.array(np.bytes_("hey")), False),
        ("hey", False),
        (np.array([np.bytes_("hey"), np.bytes_("hey")]), True),
        (5, True),
    ],
)
def test_np2str(input_value, exp_except):
    """Test the np2str function on different inputs."""
    if exp_except:
        with pytest.raises(ValueError):
            np2str(input_value)
    else:
        assert np2str(input_value) == "hey"


@pytest.mark.parametrize(
    ("input_value", "exp_result"),
    [
        (b"Hello", "Hello"),
        ("Hello", "Hello"),
        (np.bytes_("Hello"), "Hello"),
        (np.array(np.bytes_("Hello")), np.array(np.bytes_("Hello"))),
    ]
)
def test_bytes2string(input_value, exp_result):
    """Test the bytes2string function with different inputs."""
    assert bytes2string(input_value) == exp_result


def test_download_rsr_dry_run():
    """Test dry run of the download_rsr function."""
    with unittest.mock.patch("pyspectral.utils.requests") as requests_mock:
        utils.download_rsr(dry_run=True)
    requests_mock.assert_not_called()


def test_download_rsr(tmp_path):
    """Test basic usage of the download_rsr function."""
    tar_data = _create_fake_rsr_tarball_bytes()
    with responses.RequestsMock() as rsps:
        rsps.add(
            "GET",
            utils.HTTP_PYSPECTRAL_RSR,
            body=tar_data,
            status=200,
            content_type="application/octet-stream",
            headers={
                "Content-Length": str(len(tar_data)),
            },
        )
        utils.download_rsr(dest_dir=str(tmp_path))
    assert os.path.isfile(tmp_path / "PYSPECTRAL_RSR_VERSION")


def _create_fake_rsr_tarball_bytes():
    file_obj = BytesIO()

    with tarfile.open(mode="w:gz", fileobj=file_obj) as rsr_tar:
        info = tarfile.TarInfo(utils.RSR_DATA_VERSION_FILENAME)
        info.size = len(utils.RSR_DATA_VERSION)
        rsr_tar.addfile(info, BytesIO(utils.RSR_DATA_VERSION.encode("ascii")))

    file_obj.seek(0)
    return file_obj.read()


def test_download_luts_dry_run():
    """Test a dry run of the download_luts function."""
    with unittest.mock.patch("pyspectral.utils.requests") as requests_mock:
        utils.download_luts(dry_run=True)
    requests_mock.assert_not_called()


@pytest.mark.parametrize(
    ("aerosol_types", "aerosol_type", "exp_warning"),
    [
        (None, None, False),
        (["desert_aerosol"], None, False),
        (None, "desert_aerosol", True),
    ]
)
def test_download_luts(tmp_path, aerosol_types, aerosol_type, exp_warning):
    """Test basic usage of the download_luts function."""
    atypes_to_create = _atypes_to_create(aerosol_types, aerosol_type)
    tar_data = _create_fake_lut_tarball_bytes(aerosol_types=atypes_to_create)
    with responses.RequestsMock() as rsps:
        rsps.add(
            "GET",
            re.compile(re.escape(utils.LUT_URL_PREFIX) + r"_.+\.tgz"),
            body=tar_data,
            status=200,
            content_type="application/octet-stream",
            headers={
                "Content-Length": str(len(tar_data)),
            },
        )
        with _fake_get_config(tmp_path), warnings.catch_warnings(record=True) as warns:
            utils.download_luts(aerosol_types=aerosol_types, aerosol_type=aerosol_type)
    _check_user_warning(warns, exp_warning)
    _check_expected_aerosol_files(atypes_to_create, tmp_path)


def _atypes_to_create(aerosol_types, aerosol_type):
    if aerosol_types:
        return aerosol_types
    if aerosol_type is None:
        return utils.AEROSOL_TYPES
    return [aerosol_type]


def _create_fake_lut_tarball_bytes(aerosol_types):
    file_obj = BytesIO()

    with tarfile.open(mode="w:gz", fileobj=file_obj) as rsr_tar:
        for aerosol_type in aerosol_types:
            version_fn = utils.ATM_CORRECTION_LUT_VERSION[aerosol_type]["filename"]
            version_value = utils.ATM_CORRECTION_LUT_VERSION[aerosol_type]["version"]
            info = tarfile.TarInfo(version_fn)
            info.size = len(version_value)
            rsr_tar.addfile(info, BytesIO(version_value.encode("ascii")))

    file_obj.seek(0)
    return file_obj.read()


@contextlib.contextmanager
def _fake_get_config(tmp_path):
    def _get_config():
        return {
            "rayleigh_dir": str(tmp_path),
            "rsr_dir": str(tmp_path),
        }
    with unittest.mock.patch("pyspectral.utils.get_config") as get_config:
        get_config.side_effect = _get_config
        yield


def _check_user_warning(warns, exp_warning):
    all_user_warnings = [warn_msg for warn_msg in warns if issubclass(warn_msg.category, UserWarning)]
    all_matching_warnings = [warn_msg for warn_msg in all_user_warnings if "aerosol_type" in str(warn_msg.message)]
    assert len(all_matching_warnings) == (1 if exp_warning else 0)


def _check_expected_aerosol_files(atypes_to_create, tmp_path):
    for aerosol_type in utils.AEROSOL_TYPES:
        atype_fn = tmp_path / aerosol_type / utils.ATM_CORRECTION_LUT_VERSION[aerosol_type]["filename"]
        if aerosol_type in atypes_to_create:
            assert atype_fn.is_file()
        else:
            assert not atype_fn.is_file()


@pytest.mark.parametrize(
    ("platform_name", "input_value_sensor", "exp_sensor_name"),
    [
        ('Metop-C', 'avhrr/3', 'avhrr3'),
        ('NOAA-19', 'avhrr/3', 'avhrr3'),
        ('NOAA-12', 'avhrr/2', 'avhrr2'),
        ('TIROS-N', 'avhrr/1', 'avhrr1'),
        ('Metop-C', 'avhrr-3', 'avhrr3'),
        ('NOAA-12', 'avhrr-2', 'avhrr2'),
        ('TIROS-N', 'avhrr-1', 'avhrr1'),
    ],
)
def test_check_and_adjust_instrument_name_instrument_okay(platform_name, input_value_sensor, exp_sensor_name):
    """Test the checking and adjusting of the instrument name."""
    res = check_and_adjust_instrument_name(platform_name, input_value_sensor)
    assert res == exp_sensor_name


@pytest.mark.parametrize(
    ("platform_name", "input_value_sensor", "exp_sensor_name"),
    [
        ('GOES-16', 'Idontknow', 'abi'),
        ('FY-4B', 'unknown', 'agri'),
        ('Meteosat-12', 'seviri', 'fci'),
    ],
)
def test_check_and_adjust_instrument_name_instrument_name_inconsistent(caplog,
                                                                       platform_name,
                                                                       input_value_sensor,
                                                                       exp_sensor_name):
    """Test the checking and adjusting of the instrument name."""
    with caplog.at_level(logging.WARNING):
        _ = check_and_adjust_instrument_name(platform_name, input_value_sensor)

    log_output = "Inconsistent instrument/satellite input - instrument set to {instr}".format(instr=exp_sensor_name)
    assert log_output in caplog.text


@pytest.mark.parametrize(
    ("platform_name", "input_value_sensor", "exp_sensor_name"),
    [
        ('GOES-16', 'abi', 'abi'),
        ('FY-4B', 'ghi', 'ghi'),
        ('Meteosat-12', 'fci', 'fci'),
    ],
)
def test_check_and_adjust_instrument_name_correct(platform_name, input_value_sensor, exp_sensor_name):
    """Check instrument name is not changed if correct."""
    assert check_and_adjust_instrument_name(platform_name, input_value_sensor) == exp_sensor_name


def test_are_instruments_identical():
    """Test that error is raised if non-string is passed to the function."""
    res = are_instruments_identical('thing1', 'thing1')
    assert res is True
    res = are_instruments_identical('thing1', 'thing2')
    assert res is False

    with pytest.raises(ValueError):
        are_instruments_identical(['thing1', 'thing3'], 'thing2')



#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2023 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Read solar irradiances and calculate solar flux.

Module to read solar irradiance spectra and calculate the solar flux over
various instrument bands given their relative spectral response functions
"""

import logging
from pathlib import Path

import numpy as np

LOG = logging.getLogger(__name__)

# STANDARD SPECTRA from Air Mass Zero: http://rredc.nrel.gov/solar/spectra/am0/
#    * 2000 ASTM Standard Extraterrestrial Spectrum Reference E-490-00
#      (119.5 - 1,000,000.0 nm)
TOTAL_IRRADIANCE_SPECTRUM_2000ASTM = Path(__file__).resolve().parent / "data" / "e490_00a.dat"


class SolarIrradianceSpectrum(object):
    """
    Total Top of Atmosphere (TOA) Solar Irradiance Spectrum.

    Wavelength is in units of microns (10^-6 m).
    The spectral Irradiance in the file TOTAL_IRRADIANCE_SPECTRUM_2000ASTM is
    in units of W/m^2/micron
    """

    def __init__(self, filename=TOTAL_IRRADIANCE_SPECTRUM_2000ASTM, **options):
        """Initialize the top of atmosphere solar irradiance spectrum object from file.

        By default, this will use the following spectra:
        2000 ASTM Standard Extraterrestrial Spectrum Reference E-490-00

        To use a different spectra, specify the `filename` when initialising the class.

        Input:
        filename: Filename of the solar irradiance spectrum (default: 2000 ASTM)
        options:
          dlambda:
            Delta wavelength: the step in wavelength defining the resolution on
            which to integrate/convolute.
            Default is 0.005 if 'wavespace' is 'wavelength' and 2.0 if 'wavenumber'.
          wavespace:
            It is possible to specify if the solar irradiance spectrum should
            be given in terms of wavelength (default) or in terms of
            wavenumber. If the latter is desired 'wavespace' should be set to
            'wavenumber'.

        """
        self.wavelength = None
        self.wavenumber = None
        self.irradiance = None
        self.filename = filename
        self.ipol_wavelength = None
        self.ipol_irradiance = None
        self.ipol_channel_response = None
        # Delta wavelength used when resampling the
        # spectrum to an evenly spaced grid (using interpolation)
        if 'wavespace' in options:
            self.wavespace = options['wavespace']
        else:
            self.wavespace = 'wavelength'
        if 'dlambda' in options:
            self._dlambda = options['dlambda']
        else:
            if self.wavespace == 'wavelength':
                self._dlambda = 0.005
            else:
                self._dlambda = 1. / (0.005 * 100.)

        self._load()

        if self.wavespace == 'wavenumber':
            self.convert2wavenumber()
            self.units = {'irradiance': 'mW/m^2 (cm^{-1})^{-1}',
                          'flux': 'mW/m^2'}
        else:
            self.units = {'irradiance': '$W/m^2 (1e-6*m)^{-1})',
                          'flux': 'W/m^2'}

    def convert2wavenumber(self):
        """Convert from wavelengths to wavenumber.

        Units:
          Wavelength: micro meters (1e-6 m)
          Wavenumber: cm-1

        """
        self.wavenumber = 1. / (1e-4 * self.wavelength[::-1])
        self.irradiance = (self.irradiance[::-1] *
                           self.wavelength[::-1] * self.wavelength[::-1] * 0.1)
        self.wavelength = None

    def _load(self):
        """Read the tabulated spectral irradiance data from file."""
        self.wavelength, self.irradiance = \
            np.genfromtxt(self.filename, unpack=True)

    def solar_constant(self):
        """Calculate the solar constant."""
        if self.wavenumber is not None:
            return np.trapz(self.irradiance, self.wavenumber)
        if self.wavelength is not None:
            return np.trapz(self.irradiance, self.wavelength)

        raise TypeError('Neither wavelengths nor wavenumbers available!')

    def inband_solarflux(self, rsr, scale=1.0, **options):
        """Get the in band solar flux.

        Derive the inband solar flux for a given instrument relative
        spectral response valid for an earth-sun distance of one AU.
        """
        return self._band_calculations(rsr, True, scale, **options)

    def inband_solarirradiance(self, rsr, scale=1.0, **options):
        """Get the in band solar irradiance.

        Derive the inband solar irradiance for a given instrument relative
        spectral response valid for an earth-sun distance of one AU.

        (Same as the in band solar flux).
        """
        return self._band_calculations(rsr, False, scale, **options)

    def _band_calculations(self, rsr, flux, scale, **options):
        """Derive in band solar flux.

        Derive the in band solar flux or in band solar irradiance for a given
        instrument relative spectral response valid for an earth-sun distance
        of one AU.

        rsr: Relative Spectral Response (one detector only)
        Dictionary with two members 'wavelength' and 'response'
        options:
        detector: Detector number (between 1 and N - N=number of detectors
        for channel)

        """
        from scipy.interpolate import InterpolatedUnivariateSpline

        if 'detector' in options:
            detector = options['detector']
        else:
            detector = 1

        # Resample/Interpolate the response curve:
        if self.wavespace == 'wavelength':
            if 'response' in rsr:
                wvl = rsr['wavelength'] * scale
                resp = rsr['response']
            else:
                wvl = rsr['det-{0:d}'.format(detector)]['wavelength'] * scale
                resp = rsr['det-{0:d}'.format(detector)]['response']
        else:
            if 'response' in rsr:
                wvl = rsr['wavenumber'] * scale
                resp = rsr['response']
            else:
                wvl = rsr['det-{0:d}'.format(detector)]['wavenumber'] * scale
                resp = rsr['det-{0:d}'.format(detector)]['response']

        start = wvl[0]
        end = wvl[-1]
        # print "Start and end: ", start, end
        LOG.debug("Begin and end wavelength/wavenumber: %f %f ", start, end)
        dlambda = self._dlambda
        xspl = np.linspace(start, end, int(round((end - start) / self._dlambda)) + 1)

        ius = InterpolatedUnivariateSpline(wvl, resp)
        resp_ipol = ius(xspl)

        # Interpolate solar spectrum to specified resolution and over specified
        # Spectral interval:
        self.interpolate(dlambda=dlambda, ival_wavelength=(start, end))

        # Mask out outside the response curve:
        maskidx = np.logical_and(np.greater_equal(self.ipol_wavelength, start),
                                 np.less_equal(self.ipol_wavelength, end))
        wvl = np.repeat(self.ipol_wavelength, maskidx)
        irr = np.repeat(self.ipol_irradiance, maskidx)

        # Calculate the solar-flux: (w/m2)
        if flux:
            return np.trapz(irr * resp_ipol, wvl)

        # Divide by the equivalent band width:
        return np.trapz(irr * resp_ipol, wvl) / np.trapz(resp_ipol, wvl)

    def interpolate(self, **options):
        """Interpolate Irradiance to a specified evenly spaced resolution/grid.

        This is necessary to make integration and folding (with a channel
        relative spectral response) straightforward.

        dlambda = wavelength interval in microns
        start = Start of the wavelength interval (left/lower)
        end = End of the wavelength interval (right/upper end)
        options:
        dlambda: Delta wavelength used when interpolating/resampling
        ival_wavelength: Tuple. The start and end interval in wavelength
        space, defining where to integrate/convolute the spectral response
        curve on the spectral irradiance data.

        """
        from scipy.interpolate import InterpolatedUnivariateSpline

        if 'ival_wavelength' in options:
            ival_wavelength = options['ival_wavelength']
        else:
            ival_wavelength = None

        if 'dlambda' in options:
            self._dlambda = options['dlambda']

        if ival_wavelength is None:
            if self.wavespace == 'wavelength':
                start = self.wavelength[0]
                end = self.wavelength[-1]
            else:
                start = self.wavenumber[0]
                end = self.wavenumber[-1]
        else:
            start, end = ival_wavelength

        xspl = np.linspace(start, end, int(round((end - start) / self._dlambda)) + 1)
        if self.wavespace == 'wavelength':
            ius = InterpolatedUnivariateSpline(
                self.wavelength, self.irradiance)
        else:
            ius = InterpolatedUnivariateSpline(
                self.wavenumber, self.irradiance)
        yspl = ius(xspl)

        self.ipol_wavelength = xspl
        self.ipol_irradiance = yspl

    def plot(self, plotname=None, **options):
        """Plot the data."""
        if 'color' in options:
            color = options['color']
        else:
            color = 'blue'

        if self.wavespace == "wavelength":
            xlabel = r"Wavelength ($\mu m$)"
            ylabel = r"Irradiance ($W/(m^2 \mu m$))"
            xlim = [0, 4.2]
            xwl, yir = self.wavelength, self.irradiance
        elif self.wavespace == "wavenumber":
            xlabel = "Wavenumber ($cm^{-1}$)"
            ylabel = "Irradiance ($mW/m^2 (cm^{-1})^{-1}$))"
            xlim = [0, 35000]
            xwl, yir = self.wavenumber, self.irradiance
        else:
            raise TypeError('Neither wavelengths nor wavenumbers available!')

        from matplotlib import pyplot as plt
        from matplotlib import rcParams
        rcParams['text.usetex'] = True
        rcParams['text.latex.unicode'] = True

        fig, axl = plt.subplots(figsize=(8, 4))
        plot_title = "Solar Irradiance Spectrum"
        axl.set_title(plot_title)

        axl.plot(xwl, yir, '-', color=color)

        axl.set_xlabel(xlabel)
        axl.set_ylabel(ylabel)
        axl.set_xlim(xlim)
        axl.set_ylim([0, yir.max()])
        axl.grid(True)

        if plotname is None:
            plt.show()
        else:
            fig.savefig(plotname)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2013-2022 Pytroll
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Setup for the Pyspectral package."""

import os.path

from setuptools import find_packages, setup

description = ('Reading and manipulaing satellite sensor spectral responses and the '
               'solar spectrum, to perfom various corrections to VIS and NIR band data')

try:
    with open('./README.md', 'r') as fd:
        long_description = fd.read()
except IOError:
    long_description = ''

requires = ['docutils>=0.3', 'numpy', 'scipy', 'python-geotiepoints>=1.1.1',
            'h5py>=2.5', 'requests', 'pyyaml', 'platformdirs']

dask_extra = ['dask[array]']
test_requires = ['pyyaml', 'dask[array]', 'xlrd', 'pytest', 'xarray', 'responses']

NAME = 'pyspectral'

setup(name=NAME,
      description=description,
      author='Adam Dybbroe',
      author_email='adam.dybbroe@smhi.se',
      classifiers=['Development Status :: 4 - Beta',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: GNU General Public License v3 ' +
                   'or later (GPLv3+)',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python',
                   'Topic :: Scientific/Engineering'],
      url='https://github.com/pytroll/pyspectral',
      long_description=long_description,
      long_description_content_type="text/markdown",
      license='GPLv3',
      packages=find_packages(),
      include_package_data=True,
      package_data={
          # If any package contains *.txt files, include them:
          '': ['*.txt', '*.det'],
          'pyspectral': [os.path.join('etc', 'pyspectral.yaml'),
                         os.path.join('data', '*.dat'),
                         os.path.join('data', '*.XLS'),
                         'data/modis/terra/Reference_RSR_Dataset/*.det'],

      },

      # Project should use reStructuredText, so ensure that the docutils get
      # installed or upgraded on the target machine
      install_requires=requires,
      extras_require={'xlrd': ['xlrd'], 'trollsift': ['trollsift'],
                      'matplotlib': ['matplotlib'],
                      'pandas': ['pandas'],
                      'tqdm': ['tqdm'],
                      'dask': dask_extra},
      scripts=['bin/plot_rsr.py', 'bin/composite_rsr_plot.py',
               'bin/download_atm_correction_luts.py',
               'bin/download_rsr.py'],
      data_files=[('share', ['pyspectral/data/e490_00a.dat',
                             'pyspectral/data/MSG_SEVIRI_Spectral_Response_Characterisation.XLS'])],
      test_suite='pyspectral.tests.suite',
      tests_require=test_requires,
      python_requires='>=3.10',
      zip_safe=False,
      )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unittest for the rayleigh correction utilities."""
import contextlib
import os
import unittest
from unittest.mock import patch

import dask.array as da
import h5py
import numpy as np
import pytest

from pyspectral import rayleigh, utils
from pyspectral.tests.data import (
    TEST_RAYLEIGH_AZID_COORD,
    TEST_RAYLEIGH_LUT,
    TEST_RAYLEIGH_SATZ_COORD,
    TEST_RAYLEIGH_SUNZ_COORD,
    TEST_RAYLEIGH_WVL_COORD,
)
from pyspectral.utils import ATM_CORRECTION_LUT_VERSION

TEST_RAYLEIGH_RESULT1 = np.array([10.339923,    8.64748], dtype='float32')
TEST_RAYLEIGH_RESULT2 = np.array([9.653559, 8.464997], dtype='float32')
TEST_RAYLEIGH_RESULT3 = np.array([5.488735, 8.533125], dtype='float32')
TEST_RAYLEIGH_RESULT4 = np.array([0.0,    8.64748], dtype='float32')
TEST_RAYLEIGH_RESULT5 = np.array([9.653559, np.nan], dtype='float32')
TEST_RAYLEIGH_RESULT_R1 = np.array([16.66666667, 20.83333333, 25.], dtype='float32')
TEST_RAYLEIGH_RESULT_R2 = np.array([0., 6.25, 12.5], dtype='float32')

TEST_ZENITH_ANGLES_RESULTS = np.array([68.67631374, 68.67631374, 32., 0.])


class RelativeSpectralResponseTestData(object):
    """Create the class instance to hold the RSR test data."""

    def __init__(self):
        """Make a test data set of relative spectral responses."""
        self.rsr = {}
        channel_names = ['ch12', 'ch13', 'ch10', 'ch11', 'ch16', 'ch14',
                         'ch15', 'ch1', 'ch2', 'ch3', 'ch4', 'ch5', 'ch6',
                         'ch7', 'ch8', 'ch9']
        wvl = [9.6372744012936646, 10.407492196078628, 7.3468642293967275,
               8.5926867614178715, 13.280724258676756, 11.239642285822033,
               12.380741429961382, 0.47063607733748003, 0.5099976405799187,
               0.63914891611559055, 0.85668832355426627, 1.6100814361999056,
               2.2568056299864101, 3.8853663735353847, 6.2428987228916233,
               6.9411756334211789]
        ch3_wvl = np.array([0.55518544, 0.56779468, 0.58099002, 0.59481323, 0.60931027,
                            0.62453163, 0.64053291, 0.65737575, 0.67512828, 0.69386619,
                            0.71367401])
        ch3_resp = np.array([2.61000005e-05, 1.07899999e-04, 3.26119992e-03,
                             2.90650606e-01, 9.02460396e-01, 9.60878074e-01,
                             9.97266889e-01, 9.94823873e-01, 7.18220174e-01,
                             8.31819978e-03, 9.34999989e-05])

        idx = 0
        for chname in channel_names:
            self.rsr[chname] = {'det-1': {}}
            self.rsr[chname]['det-1']['central_wavelength'] = wvl[idx]
            idx = idx + 1

        chname = 'ch3'
        self.rsr[chname]['det-1']['wavelength'] = ch3_wvl
        self.rsr[chname]['det-1']['response'] = ch3_resp


@pytest.fixture
def fake_lut_hdf5(tmp_path):
    """Create a fake LUT HDF5 file and necessary mocks to use it."""
    aerosol_type = "marine_clean_aerosol"
    base_dir = tmp_path / aerosol_type
    base_dir.mkdir()
    _create_fake_lut_version_file(base_dir, aerosol_type)
    for atm_type in ("midlatitude_summer", "subarctic_winter", "tropical", "us-standard"):
        _create_fake_lut_hdf5_file(base_dir, atm_type)
    fake_config = {
        "rsr_dir": str(tmp_path),
        "rayleigh_dir": str(tmp_path),
        "download_from_internet": False,
    }
    with patch('pyspectral.rayleigh.get_config', lambda: fake_config), \
            patch('pyspectral.utils.get_config', lambda: fake_config):
        yield None


def _create_fake_lut_version_file(base_dir, aerosol_type):
    lut_version = ATM_CORRECTION_LUT_VERSION[aerosol_type]["version"]
    lutfiles_path = str(base_dir / ATM_CORRECTION_LUT_VERSION[aerosol_type]["filename"])
    with open(lutfiles_path, "w") as version_file:
        version_file.write(lut_version)


def _create_fake_lut_hdf5_file(base_dir, atm_type) -> str:
    filename = str(base_dir / f"rayleigh_lut_{atm_type}.h5")
    with h5py.File(filename, "w") as h5f:
        h5f["reflectance"] = TEST_RAYLEIGH_LUT
        h5f["wavelengths"] = TEST_RAYLEIGH_WVL_COORD
        h5f["azimuth_difference"] = TEST_RAYLEIGH_AZID_COORD
        h5f["sun_zenith_secant"] = TEST_RAYLEIGH_SUNZ_COORD
        h5f["satellite_zenith_secant"] = TEST_RAYLEIGH_SATZ_COORD
    return filename


def _create_rayleigh(platform='NOAA-20', sensor='VIIRS'):
    rayl = rayleigh.Rayleigh(platform, sensor, atmosphere='midlatitude summer')
    return rayl


@contextlib.contextmanager
def mocked_rsr():
    """Mock the RSR class used by the Rayleigh class with fake data."""
    with patch('pyspectral.rayleigh.RelativeSpectralResponse') as mymock:
        instance = mymock.return_value
        instance.rsr = RelativeSpectralResponseTestData().rsr
        instance.unit = '1e-6 m'
        instance.si_scale = 1e-6
        yield mymock


def _create_dask_array(input_data, dtype):
    return da.from_array(np.array(input_data, dtype=dtype))


class TestRayleighDask:
    """Class for testing pyspectral.rayleigh - with dask-arrays as input."""

    def test_dask_clipping_angles_with_nans(self):
        """Test the clipping of angles outside coordinate range - with nan's in input."""
        from pyspectral.rayleigh import _clip_angles_inside_coordinate_range
        zenith_angle = da.array([79., 69., 32., np.nan])
        result = _clip_angles_inside_coordinate_range(zenith_angle, 2.75)
        np.testing.assert_allclose(result, TEST_ZENITH_ANGLES_RESULTS)

    def test_get_reflectance_dask_redband_outside_clip(self, fake_lut_hdf5):
        """Test getting the reflectance correction with dask inputs - using red band reflections outside 20-100."""
        sun_zenith = da.array([67., 32.])
        sat_zenith = da.array([45., 18.])
        azidiff = da.array([150., 110.])
        redband_refl = da.array([108., -0.5])
        viirs_rayl = _create_rayleigh()
        with mocked_rsr():
            refl_corr = viirs_rayl.get_reflectance(sun_zenith, sat_zenith, azidiff, 'ch3', redband_refl)
        np.testing.assert_allclose(refl_corr, TEST_RAYLEIGH_RESULT4)
        assert isinstance(refl_corr, da.Array)

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    @pytest.mark.parametrize("use_dask", [False, True])
    @pytest.mark.parametrize(
        ("input_data", "exp_result"),
        [
            (([67.0, 32.0], [45.0, 18.0], [150.0, 110.0], [14.0, 5.0]), TEST_RAYLEIGH_RESULT1),
            (([60.0, 20.0], [49.0, 26.0], [140.0, 130.0], [12.0, 8.0]), TEST_RAYLEIGH_RESULT2),
        ]
    )
    def test_get_reflectance(self, fake_lut_hdf5, dtype, use_dask, input_data, exp_result):
        """Test getting the reflectance correction with dask inputs."""
        array_func = np.array if not use_dask else _create_dask_array
        sun_zenith = array_func(input_data[0], dtype=dtype)
        sat_zenith = array_func(input_data[1], dtype=dtype)
        azidiff = array_func(input_data[2], dtype=dtype)
        redband_refl = array_func(input_data[3], dtype=dtype)
        rayl = _create_rayleigh()
        with mocked_rsr():
            refl_corr = rayl.get_reflectance(sun_zenith, sat_zenith, azidiff, 'ch3', redband_refl)

        if use_dask:
            assert isinstance(refl_corr, da.Array)
            refl_corr_np = refl_corr.compute()
            assert refl_corr_np.dtype == refl_corr.dtype  # check that the final numpy array's dtype is equal
            refl_corr = refl_corr_np

        assert isinstance(refl_corr, np.ndarray)
        np.testing.assert_allclose(refl_corr, exp_result.astype(dtype), atol=4.0e-06)
        assert refl_corr.dtype == dtype  # check that the dask array's dtype is equal

    def test_get_reflectance_wvl_outside_range(self, fake_lut_hdf5):
        """Test getting the reflectance correction with wavelength outside correction range."""
        with mocked_rsr() as rsr_obj:
            rsr_obj.side_effect = IOError("No rsr data in pyspectral for this platform and sensor")
            sun_zenith = da.from_array([50., 10.])
            sat_zenith = da.from_array([39., 16.])
            azidiff = da.from_array([170., 110.])
            redband_refl = da.from_array([29., 12.])
            rayl = rayleigh.Rayleigh('UFO', 'unknown', atmosphere='midlatitude summer')

            # we gave a direct wavelength so the RSR object shouldn't be needed
            refl_corr = rayl.get_reflectance(sun_zenith, sat_zenith, azidiff, 1.2, redband_refl)
            np.testing.assert_allclose(refl_corr, 0)
            assert isinstance(refl_corr, da.Array)
            rsr_obj.assert_not_called()


class TestRayleigh:
    """Class for testing pyspectral.rayleigh."""

    def test_get_effective_wavelength_and_band_name_with_floats(self, fake_lut_hdf5):
        """Test getting the effective wavelength."""
        this = rayleigh.Rayleigh('Himawari-8', 'ahi')
        # Only ch3 (~0.63) testdata implemented yet...
        ewl, band_name = this._get_effective_wavelength_and_band_name(0.65)
        np.testing.assert_allclose(ewl, 650)  # 635.6167)
        assert isinstance(band_name, str)

        this = rayleigh.Rayleigh('Himawari-8', 'ahi')
        ewl, band_name = this._get_effective_wavelength_and_band_name(0.7)
        assert ewl == 700.0
        assert isinstance(band_name, str)
        ewl, band_name = this._get_effective_wavelength_and_band_name(0.9)
        assert ewl == 900.0
        assert isinstance(band_name, str)
        ewl, band_name = this._get_effective_wavelength_and_band_name(0.455)
        assert ewl == 455.0
        assert isinstance(band_name, str)

    def test_rayleigh_init(self, fake_lut_hdf5):
        """Test creating the Rayleigh object."""
        with patch('pyspectral.rayleigh.RelativeSpectralResponse') as mymock:
            instance = mymock.return_value
            instance.rsr = RelativeSpectralResponseTestData().rsr

            with pytest.raises(AttributeError):
                rayleigh.Rayleigh('Himawari-8', 'ahi', atmosphere='unknown')

            with pytest.raises(AttributeError):
                rayleigh.Rayleigh('Himawari-8', 'ahi', aerosol_type='unknown')

            this = rayleigh.Rayleigh('Himawari-8', 'ahi', atmosphere='subarctic winter')
            assert os.path.basename(this.reflectance_lut_filename).endswith('subarctic_winter.h5')
            assert this.sensor == 'ahi'

            this = rayleigh.Rayleigh('NOAA-19', 'avhrr/3', atmosphere='tropical')
            assert this.sensor == 'avhrr3'

    def test_clipping_angles_with_nans(self):
        """Test the clipping of angles outside coordinate range - with nan's in input."""
        from pyspectral.rayleigh import _clip_angles_inside_coordinate_range
        zenith_angle = np.array([79., 69., 32., np.nan])
        result = _clip_angles_inside_coordinate_range(zenith_angle, 2.75)
        np.testing.assert_allclose(result, TEST_ZENITH_ANGLES_RESULTS)

    def test_rayleigh_reduction(self, fake_lut_hdf5):
        """Test the code that reduces Rayleigh correction for high zenith angles."""
        # Test the Rayleigh reduction code
        sun_zenith = np.array([70., 65., 60.])
        in_rayleigh = [50, 50, 50]
        # Test case where no reduction is done.
        retv = rayleigh.Rayleigh.reduce_rayleigh_highzenith(sun_zenith, in_rayleigh, 70., 90., 1)
        np.testing.assert_allclose(retv, in_rayleigh)
        # Test case where moderate reduction is performed.
        retv = rayleigh.Rayleigh.reduce_rayleigh_highzenith(sun_zenith, in_rayleigh, 30., 90., 1)
        np.testing.assert_allclose(retv, TEST_RAYLEIGH_RESULT_R1)
        # Test case where extreme reduction is performed.
        retv = rayleigh.Rayleigh.reduce_rayleigh_highzenith(sun_zenith, in_rayleigh, 30., 90., 1.5)
        np.testing.assert_allclose(retv, TEST_RAYLEIGH_RESULT_R2)

    def test_rayleigh_getname(self):
        """Test logic for Rayleigh instrument selection."""
        with pytest.raises(ValueError):
            _create_rayleigh(platform='FY-4B')

        rayl = _create_rayleigh(platform='FY-4A', sensor='agri')
        assert rayl.sensor == 'agri'

        rayl = _create_rayleigh(platform='FY-4B', sensor='agri')
        assert rayl.sensor == 'agri'

        rayl = _create_rayleigh(platform='FY-4B', sensor='ghi')
        assert rayl.sensor == 'ghi'

        with pytest.raises(ValueError):
            _create_rayleigh(platform='FY-4B', sensor='nosensor')

    @patch('pyspectral.rayleigh.da', None)
    def test_get_reflectance_redband_outside_clip(self, fake_lut_hdf5):
        """Test getting the reflectance correction - using red band reflections outside 20 to 100."""
        sun_zenith = np.array([67., 32.])
        sat_zenith = np.array([45., 18.])
        azidiff = np.array([150., 110.])
        redband_refl = np.array([100., 20.])
        rayl = _create_rayleigh()
        with mocked_rsr():
            refl_corr1 = rayl.get_reflectance(
                sun_zenith, sat_zenith, azidiff, 'ch3', redband_refl)

        np.testing.assert_allclose(refl_corr1, TEST_RAYLEIGH_RESULT4)

        rng = np.random.default_rng(12345)
        rints_low = rng.integers(low=-10, high=20, size=2).astype('float')
        rints_high = rng.integers(low=100, high=200, size=2).astype('float')

        redband_refl = np.array([rints_high[0], rints_low[0]])
        with mocked_rsr():
            refl_corr2 = rayl.get_reflectance(
                sun_zenith, sat_zenith, azidiff, 'ch3', redband_refl)

        redband_refl = np.array([rints_high[1], rints_low[1]])
        with mocked_rsr():
            refl_corr3 = rayl.get_reflectance(
                sun_zenith, sat_zenith, azidiff, 'ch3', redband_refl)

        assert isinstance(refl_corr1, np.ndarray)
        assert isinstance(refl_corr2, np.ndarray)
        np.testing.assert_allclose(refl_corr1, refl_corr2)
        np.testing.assert_allclose(refl_corr2, refl_corr3)

    @patch('pyspectral.rayleigh.da', None)
    @pytest.mark.parametrize(
        ("sun_zenith", "sat_zenith", "azidiff", "redband_refl", "exp_result"),
        [
            (np.array([67., 32.]), np.array([45., 18.]), np.array([150., 110.]), np.array([14., 5.]),
             TEST_RAYLEIGH_RESULT1),
            (np.array([60., 20.]), np.array([49., 26.]), np.array([140., 130.]), np.array([12., 8.]),
             TEST_RAYLEIGH_RESULT2),
            (np.array([60., 20.]), np.array([49., 26.]), np.array([140., 130.]), np.array([12., np.nan]),
             TEST_RAYLEIGH_RESULT5),
        ]
    )
    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    def test_get_reflectance(self, fake_lut_hdf5, sun_zenith, sat_zenith, azidiff, redband_refl, exp_result, dtype):
        """Test getting the reflectance correction."""
        rayl = _create_rayleigh()
        with mocked_rsr():
            refl_corr = rayl.get_reflectance(
                sun_zenith.astype(dtype),
                sat_zenith.astype(dtype),
                azidiff.astype(dtype),
                'ch3',
                redband_refl.astype(dtype))
        assert isinstance(refl_corr, np.ndarray)
        np.testing.assert_allclose(refl_corr, exp_result.astype(dtype), atol=4.0e-06)

    @patch('pyspectral.rayleigh.da', None)
    def test_get_reflectance_no_rsr(self, fake_lut_hdf5):
        """Test getting the reflectance correction, simulating that we have no RSR data."""
        with mocked_rsr() as rsr_obj:
            rsr_obj.side_effect = IOError("No rsr data in pyspectral for this platform and sensor")
            sun_zenith = np.array([50., 10.])
            sat_zenith = np.array([39., 16.])
            azidiff = np.array([170., 110.])
            redband_refl = np.array([29., 12.])
            ufo = rayleigh.Rayleigh('UFO', 'unknown', atmosphere='midlatitude summer')
            with pytest.raises(KeyError):
                ufo.get_reflectance(sun_zenith, sat_zenith, azidiff, 'ch3', redband_refl)

    @patch('pyspectral.rayleigh.da', None)
    def test_get_reflectance_float_wavelength(self, fake_lut_hdf5):
        """Test getting the reflectance correction."""
        with mocked_rsr() as rsr_obj:
            rsr_obj.side_effect = IOError("No rsr data in pyspectral for this platform and sensor")
            sun_zenith = np.array([50., 10.])
            sat_zenith = np.array([39., 16.])
            azidiff = np.array([170., 110.])
            redband_refl = np.array([29., 12.])
            # rayl = _create_rayleigh()
            rayl = rayleigh.Rayleigh('UFO', 'unknown', atmosphere='midlatitude summer')

            # we gave a direct wavelength so the RSR object shouldn't be needed
            refl_corr = rayl.get_reflectance(sun_zenith, sat_zenith, azidiff, 0.634, redband_refl)
            np.testing.assert_allclose(refl_corr, TEST_RAYLEIGH_RESULT3)
            assert isinstance(refl_corr, np.ndarray)
            rsr_obj.assert_not_called()

    @patch('pyspectral.rayleigh.da', None)
    def test_get_reflectance_wvl_outside_range(self, fake_lut_hdf5):
        """Test getting the reflectance correction with wavelength outside correction range."""
        with mocked_rsr() as rsr_obj:
            rsr_obj.side_effect = IOError("No rsr data in pyspectral for this platform and sensor")
            sun_zenith = np.array([50., 10.])
            sat_zenith = np.array([39., 16.])
            azidiff = np.array([170., 110.])
            redband_refl = np.array([29., 12.])
            rayl = rayleigh.Rayleigh('UFO', 'unknown', atmosphere='midlatitude summer')

            # we gave a direct wavelength so the RSR object shouldn't be needed
            refl_corr = rayl.get_reflectance(sun_zenith, sat_zenith, azidiff, 1.2, redband_refl)
            np.testing.assert_allclose(refl_corr, 0)
            assert isinstance(refl_corr, np.ndarray)
            rsr_obj.assert_not_called()

    def test_get_reflectance_no_lut(self, fake_lut_hdf5):
        """Test that missing a LUT causes an exception.."""
        # test LUT doesn't have a subartic_summer file
        with pytest.raises(IOError):
            rayleigh.Rayleigh('UFO', 'unknown', atmosphere='subarctic summer')


@pytest.mark.parametrize(
    ("version", "exp_download"),
    [
        (None, False),
        ("v0.0.0", True),
    ],
)
def test_check_and_download(tmp_path, version, exp_download):
    """Test that check_and_download only downloads when necessary."""
    from pyspectral.rayleigh import check_and_download
    with _fake_lut_dir(tmp_path, version), unittest.mock.patch("pyspectral.rayleigh.download_luts") as download:
        check_and_download()
        if exp_download:
            download.assert_called()
        else:
            download.assert_not_called()


@contextlib.contextmanager
def _fake_lut_dir(tmp_path, lut_version):
    with _fake_get_config(tmp_path):
        for aerosol_type in utils.AEROSOL_TYPES:
            atype_version_fn = utils.ATM_CORRECTION_LUT_VERSION[aerosol_type]["filename"]
            atype_version = utils.ATM_CORRECTION_LUT_VERSION[aerosol_type]["version"]
            atype_subdir = tmp_path / aerosol_type
            atype_subdir.mkdir()
            version_filename = str(atype_subdir / atype_version_fn)
            with open(version_filename, "w") as version_file:
                version_file.write(lut_version or atype_version)
        yield


@contextlib.contextmanager
def _fake_get_config(tmp_path):
    def _get_config():
        return {
            "rayleigh_dir": str(tmp_path),
            "rsr_dir": str(tmp_path),
            "download_from_internet": True,
        }
    with unittest.mock.patch("pyspectral.rayleigh.get_config") as get_config, \
            unittest.mock.patch("pyspectral.utils.get_config") as get_config2:
        get_config.side_effect = _get_config
        get_config2.side_effect = _get_config
        yield



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2012, 2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit testing the Blackbody/Plack radiation derivation."""

import dask
import dask.array as da
import numpy as np
import pytest

from pyspectral.blackbody import blackbody, blackbody_rad2temp, blackbody_wn, blackbody_wn_rad2temp
from pyspectral.tests.unittest_helpers import ComputeCountingScheduler

RAD_11MICRON_300KELVIN = 9573176.935507433
RAD_11MICRON_301KELVIN = 9714686.576498277

# Radiances in wavenumber space (SI-units)
WN_RAD_11MICRON_300KELVIN = 0.00115835441353
WN_RAD_11MICRON_301KELVIN = 0.00117547716523

__unittest = True


class TestBlackbody:
    """Unit testing the blackbody function."""

    def test_blackbody(self):
        """Calculate the blackbody radiation from wavelengths and temperatures."""
        wavel = 11. * 1E-6
        black = blackbody(wavel, 300.)
        assert black == pytest.approx(RAD_11MICRON_300KELVIN)
        temp1 = blackbody_rad2temp(wavel, black)
        assert temp1 == pytest.approx(300.0, 4)

        black = blackbody(wavel, 301.)
        assert black == pytest.approx(RAD_11MICRON_301KELVIN)
        temp2 = blackbody_rad2temp(wavel, black)
        assert temp2 == pytest.approx(301.0, 4)

        tb_therm = np.array([[300., 301], [299, 298], [279, 286]])
        black = blackbody(10. * 1E-6, tb_therm)
        assert isinstance(black, np.ndarray)

    def test_blackbody_dask_wave_tuple(self):
        """Calculate the blackbody radiation from wavelengths and temperatures with dask arrays."""
        tb_therm = da.array([[300., 301], [299, 298], [279, 286]])
        with dask.config.set(scheduler=ComputeCountingScheduler(0)):
            black = blackbody(10. * 1E-6, tb_therm)
        assert isinstance(black, da.Array)
        assert black.dtype == np.float64

    @pytest.mark.parametrize("dtype", (np.float32, np.float64, float))
    def test_blackbody_dask_wave_array(self, dtype):
        """Test blackbody calculations with dask arrays as inputs."""
        tb_therm = da.array([[300., 301], [0., 298], [279, 286]], dtype=dtype)
        with dask.config.set(scheduler=ComputeCountingScheduler(0)):
            black = blackbody(da.array([10. * 1E-6, 11.e-6], dtype=dtype), tb_therm)
        assert isinstance(black, da.Array)
        assert black.dtype == dtype

    def test_blackbody_wn(self):
        """Calculate the blackbody radiation from wavenumbers and temperatures."""
        wavenumber = 90909.1  # 11 micron band
        black = blackbody_wn(wavenumber, 300.)
        assert black == pytest.approx(WN_RAD_11MICRON_300KELVIN)
        temp1 = blackbody_wn_rad2temp(wavenumber, black)
        assert temp1 == pytest.approx(300.0, 4)

        black = blackbody_wn(wavenumber, 301.)
        assert black == pytest.approx(WN_RAD_11MICRON_301KELVIN)
        temp2 = blackbody_wn_rad2temp(wavenumber, black)
        assert temp2 == pytest.approx(301.0, 4)

        t__ = blackbody_wn_rad2temp(wavenumber, np.array([0.001, 0.0009]))
        expected = [290.3276916, 283.76115441]
        np.testing.assert_allclose(t__, expected)

        radiances = np.array([0.001, 0.0009, 0.0012, 0.0018]).reshape(2, 2)
        t__ = blackbody_wn_rad2temp(wavenumber, radiances)
        expected = np.array([290.3276916, 283.76115441,
                             302.4181330, 333.1414164]).reshape(2, 2)
        np.testing.assert_allclose(t__, expected)

    def test_blackbody_wn_dask(self):
        """Test that blackbody rad2temp preserves dask arrays."""
        import dask
        import dask.array as da
        wavenumber = 90909.1  # 11 micron band
        radiances = da.from_array([0.001, 0.0009, 0.0012, 0.0018], chunks=2).reshape(2, 2)
        with dask.config.set(scheduler=ComputeCountingScheduler(0)):
            t__ = blackbody_wn_rad2temp(wavenumber, radiances)
        assert isinstance(t__, da.Array)
        t__ = t__.compute()
        expected = np.array([290.3276916, 283.76115441,
                             302.4181330, 333.1414164]).reshape(2, 2)
        np.testing.assert_allclose(t__, expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017-2022 Pytroll developers
#
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for unit testing."""

import numpy as np


def assertNumpyArraysEqual(self, other):
    """Assert if two numpy arrays are equal."""
    if self.shape != other.shape:
        raise AssertionError("Shapes don't match")
    if not np.allclose(self, other):
        raise AssertionError("Elements don't match!")


class ComputeCountingScheduler:
    """Scheduler raising an exception if data are computed too many times."""

    def __init__(self, max_computes=1):
        """Set starting and maximum compute counts."""
        self.max_computes = max_computes
        self.total_computes = 0

    def __call__(self, dsk, keys, **kwargs):
        """Compute dask task and keep track of number of times we do so."""
        import dask
        self.total_computes += 1
        if self.total_computes > self.max_computes:
            raise RuntimeError("Too many dask computations were scheduled: "
                               "{}".format(self.total_computes))
        return dask.get(dsk, keys, **kwargs)
